@@ -6,7 +6,10 @@
  */
 
 #include "xlsx_sheet_context.hpp"
+#include "xlsx_autofilter_context.hpp"
+#include "xlsx_conditional_format_context.hpp"
 #include "xlsx_session_data.hpp"
+#include "xlsx_types.hpp"
 #include "ooxml_global.hpp"
 #include "ooxml_schemas.hpp"
 #include "ooxml_token_constants.hpp"
@@ -26,7 +29,7 @@ namespace orcus {
 
 namespace {
 
-class col_attr_parser : public std::unary_function<void, xml_token_attr_t>
+class col_attr_parser : public std::unary_function<xml_token_attr_t, void>
 {
     long m_min;
     long m_max;
@@ -77,7 +80,7 @@ public:
     bool is_hidden() const { return m_hidden; }
 };
 
-class row_attr_parser : public std::unary_function<void, xml_token_attr_t>
+class row_attr_parser : public std::unary_function<xml_token_attr_t, void>
 {
     spreadsheet::row_t m_row;
     length_t m_height;
@@ -103,7 +106,7 @@ public:
             case XML_ht:
             {
                 m_height.value = to_double(attr.value);
-                m_height.unit = length_unit_point;
+                m_height.unit = length_unit_t::point;
             }
             break;
             case XML_hidden:
@@ -132,14 +135,14 @@ class cell_attr_parser : public std::unary_function<xml_token_attr_t, void>
         address(spreadsheet::row_t _row, spreadsheet::col_t _col) : row(_row), col(_col) {}
     };
 
-    xlsx_sheet_context::cell_type m_type;
+    xlsx_cell_t m_type;
     address m_address;
     size_t m_xf;
     bool m_contains_address;
 
 public:
     cell_attr_parser() :
-        m_type(xlsx_sheet_context::cell_type_value),
+        m_type(xlsx_ct_numeric),
         m_address(0,0),
         m_xf(0),
         m_contains_address(false) {}
@@ -155,7 +158,7 @@ public:
             break;
             case XML_t:
                 // cell type
-                m_type = to_cell_type(attr.value);
+                m_type = to_xlsx_cell_type(attr.value);
             break;
             case XML_s:
                 // cell style
@@ -164,7 +167,7 @@ public:
         }
     }
 
-    xlsx_sheet_context::cell_type get_cell_type() const { return m_type; }
+    xlsx_cell_t get_cell_type() const { return m_type; }
 
     spreadsheet::row_t get_row() const { return m_address.row; }
     spreadsheet::col_t get_col() const { return m_address.col; }
@@ -172,25 +175,6 @@ public:
     bool contains_address() const { return m_contains_address; }
 
 private:
-    xlsx_sheet_context::cell_type to_cell_type(const pstring& s) const
-    {
-        xlsx_sheet_context::cell_type t = xlsx_sheet_context::cell_type_value;
-        if (s == "s")
-            t = xlsx_sheet_context::cell_type_string;
-        else if (s == "str")
-            // formula string
-            t = xlsx_sheet_context::cell_type_formula_string;
-        else if (s == "b")
-            // boolean
-            t = xlsx_sheet_context::cell_type_boolean;
-        else if (s == "e")
-            // error
-            t = xlsx_sheet_context::cell_type_error;
-        else if (s == "inlineStr")
-            t = xlsx_sheet_context::cell_type_inline_string;
-
-        return t;
-    }
 
     address to_cell_address(const pstring& s) const
     {
@@ -242,11 +226,11 @@ public:
             case XML_t:
             {
                 if (attr.value == "shared")
-                    m_attrs.type = spreadsheet::formula_shared;
+                    m_attrs.type = spreadsheet::formula_t::shared;
                 else if (attr.value == "array")
-                    m_attrs.type = spreadsheet::formula_array;
+                    m_attrs.type = spreadsheet::formula_t::array;
                 else if (attr.value == "dataTable")
-                    m_attrs.type = spreadsheet::formula_data_table;
+                    m_attrs.type = spreadsheet::formula_t::data_table;
             }
             break;
             case XML_ref:
@@ -284,7 +268,7 @@ public:
 }
 
 xlsx_sheet_context::formula::formula() :
-    type(spreadsheet::formula_normal),
+    type(spreadsheet::formula_t::normal),
     str(), ref(),
     data_table_ref1(),
     data_table_ref2(),
@@ -306,7 +290,7 @@ xlsx_sheet_context::xlsx_sheet_context(
     m_sheet_id(sheet_id),
     m_cur_row(-1),
     m_cur_col(-1),
-    m_cur_cell_type(cell_type_value),
+    m_cur_cell_type(xlsx_ct_numeric),
     m_cur_cell_xf(0)
 {
 }
@@ -317,16 +301,47 @@ xlsx_sheet_context::~xlsx_sheet_context()
 
 bool xlsx_sheet_context::can_handle_element(xmlns_id_t ns, xml_token_t name) const
 {
+    if (ns == NS_ooxml_xlsx && name == XML_autoFilter)
+        return false;
+    else if (ns == NS_ooxml_xlsx && name == XML_conditionalFormatting)
+        return false;
+
     return true;
 }
 
 xml_context_base* xlsx_sheet_context::create_child_context(xmlns_id_t ns, xml_token_t name)
 {
+    if (ns == NS_ooxml_xlsx && name == XML_autoFilter)
+    {
+        mp_child.reset(new xlsx_autofilter_context(get_session_context(), get_tokens()));
+        mp_child->transfer_common(*this);
+        return mp_child.get();
+    }
+    else if (ns == NS_ooxml_xlsx && name == XML_conditionalFormatting
+            && mp_sheet->get_conditional_format())
+    {
+        mp_child.reset(new xlsx_conditional_format_context(get_session_context(), get_tokens(),
+                    *mp_sheet->get_conditional_format()));
+        mp_child->transfer_common(*this);
+        return mp_child.get();
+    }
     return NULL;
 }
 
 void xlsx_sheet_context::end_child_context(xmlns_id_t ns, xml_token_t name, xml_context_base* child)
 {
+    if (!child)
+        return;
+
+    if (ns == NS_ooxml_xlsx && name == XML_autoFilter)
+    {
+        spreadsheet::iface::import_auto_filter* af = mp_sheet->get_auto_filter();
+        if (!af)
+            return;
+
+        const xlsx_autofilter_context& cxt = static_cast<const xlsx_autofilter_context&>(*child);
+        cxt.push_to_model(*af);
+    }
 }
 
 void xlsx_sheet_context::start_element(xmlns_id_t ns, xml_token_t name, const xml_attrs_t& attrs)
@@ -337,7 +352,8 @@ void xlsx_sheet_context::start_element(xmlns_id_t ns, xml_token_t name, const xm
     {
         case XML_worksheet:
         {
-            print_attrs(get_tokens(), attrs);
+            if (get_config().debug)
+                print_attrs(get_tokens(), attrs);
         }
         break;
         case XML_cols:
@@ -358,7 +374,7 @@ void xlsx_sheet_context::start_element(xmlns_id_t ns, xml_token_t name, const xm
                 for (spreadsheet::col_t col = func.get_min(); col <= func.get_max(); ++col)
                 {
                     if (contains_width)
-                        sheet_props->set_column_width(col-1, width, length_unit_xlsx_column_digit);
+                        sheet_props->set_column_width(col-1, width, length_unit_t::xlsx_column_digit);
                     sheet_props->set_column_hidden(col-1, hidden);
                 }
             }
@@ -385,7 +401,12 @@ void xlsx_sheet_context::start_element(xmlns_id_t ns, xml_token_t name, const xm
         }
         break;
         case XML_pageMargins:
-            xml_element_expected(parent, NS_ooxml_xlsx, XML_worksheet);
+        {
+            xml_elem_stack_t elems;
+            elems.push_back(xml_token_pair_t(NS_ooxml_xlsx, XML_worksheet));
+            elems.push_back(xml_token_pair_t(NS_ooxml_xlsx, XML_customSheetView));
+            xml_element_expected(parent, elems);
+        }
         break;
         case XML_sheetViews:
             xml_element_expected(parent, NS_ooxml_xlsx, XML_worksheet);
@@ -394,7 +415,12 @@ void xlsx_sheet_context::start_element(xmlns_id_t ns, xml_token_t name, const xm
             xml_element_expected(parent, NS_ooxml_xlsx, XML_sheetViews);
         break;
         case XML_selection:
-            xml_element_expected(parent, NS_ooxml_xlsx, XML_sheetView);
+        {
+            xml_elem_stack_t elems;
+            elems.push_back(xml_token_pair_t(NS_ooxml_xlsx, XML_sheetView));
+            elems.push_back(xml_token_pair_t(NS_ooxml_xlsx, XML_customSheetView));
+            xml_element_expected(parent, elems);
+        }
         break;
         case XML_sheetData:
             xml_element_expected(parent, NS_ooxml_xlsx, XML_worksheet);
@@ -418,7 +444,7 @@ void xlsx_sheet_context::start_element(xmlns_id_t ns, xml_token_t name, const xm
             if (sheet_props)
             {
                 length_t ht = func.get_height();
-                if (ht.unit != length_unit_unknown)
+                if (ht.unit != length_unit_t::unknown)
                     sheet_props->set_row_height(m_cur_row, ht.value, ht.unit);
 
                 bool hidden = func.is_hidden();
@@ -459,48 +485,21 @@ void xlsx_sheet_context::start_element(xmlns_id_t ns, xml_token_t name, const xm
         case XML_v:
             xml_element_expected(parent, NS_ooxml_xlsx, XML_c);
         break;
-        case XML_autoFilter:
-        {
+        case XML_tableParts:
             xml_element_expected(parent, NS_ooxml_xlsx, XML_worksheet);
-            spreadsheet::iface::import_auto_filter* af = mp_sheet->get_auto_filter();
-            if (af)
-            {
-                pstring ref = for_each(
-                    attrs.begin(), attrs.end(),
-                    single_attr_getter(m_pool, NS_ooxml_xlsx, XML_ref)).get_value();
-                af->set_range(ref.get(), ref.size());
-            }
-        }
         break;
-        case XML_filterColumn:
+        case XML_tablePart:
         {
-            xml_element_expected(parent, NS_ooxml_xlsx, XML_autoFilter);
-            spreadsheet::iface::import_auto_filter* af = mp_sheet->get_auto_filter();
-            if (af)
-            {
-                pstring colid = for_each(
-                    attrs.begin(), attrs.end(),
-                    single_attr_getter(m_pool, NS_ooxml_xlsx, XML_colId)).get_value();
-                spreadsheet::col_t col = to_long(colid);
-                af->set_column(col);
-            }
-        }
-        break;
-        case XML_filters:
-            xml_element_expected(parent, NS_ooxml_xlsx, XML_filterColumn);
-        break;
-        case XML_filter:
-        {
-            xml_element_expected(parent, NS_ooxml_xlsx, XML_filters);
-            spreadsheet::iface::import_auto_filter* af = mp_sheet->get_auto_filter();
-            if (af)
-            {
-                pstring val = for_each(
-                    attrs.begin(), attrs.end(),
-                    single_attr_getter(m_pool, NS_ooxml_xlsx, XML_val)).get_value();
-                if (!val.empty())
-                    af->append_column_match_value(val.get(), val.size());
-            }
+            xml_element_expected(parent, NS_ooxml_xlsx, XML_tableParts);
+
+            // The rid string must be pooled to the session context's string
+            // pool as it is used long after thet sheet context is deleted.
+            single_attr_getter func(get_session_context().m_string_pool, NS_ooxml_r, XML_id);
+            pstring rid = for_each(attrs.begin(), attrs.end(), func).get_value();
+
+            unique_ptr<xlsx_rel_table_info> p(new xlsx_rel_table_info);
+            p->sheet_interface = mp_sheet;
+            m_rel_extras.data.insert(opc_rel_extras_t::map_type::value_type(rid, p.release()));
         }
         break;
         default:
@@ -522,20 +521,6 @@ bool xlsx_sheet_context::end_element(xmlns_id_t ns, xml_token_t name)
         case XML_v:
             m_cur_value = m_cur_str;
         break;
-        case XML_autoFilter:
-        {
-            spreadsheet::iface::import_auto_filter* af = mp_sheet->get_auto_filter();
-            if (af)
-                af->commit();
-        }
-        break;
-        case XML_filterColumn:
-        {
-            spreadsheet::iface::import_auto_filter* af = mp_sheet->get_auto_filter();
-            if (af)
-                af->commit_column();
-        }
-        break;
         default:
             ;
     }
@@ -558,37 +543,37 @@ void xlsx_sheet_context::end_element_cell()
 
     if (!m_cur_formula.str.empty())
     {
-        if (m_cur_formula.type == spreadsheet::formula_shared && m_cur_formula.shared_id >= 0)
+        if (m_cur_formula.type == spreadsheet::formula_t::shared && m_cur_formula.shared_id >= 0)
         {
             // shared formula expression
             session_data.m_shared_formulas.push_back(
-                new xlsx_session_data::shared_formula(
+                orcus::make_unique<xlsx_session_data::shared_formula>(
                     m_sheet_id, m_cur_row, m_cur_col, m_cur_formula.shared_id,
                     m_cur_formula.str.str(), m_cur_formula.ref.str()));
         }
-        else if (m_cur_formula.type == spreadsheet::formula_array)
+        else if (m_cur_formula.type == spreadsheet::formula_t::array)
         {
             // array formula expression
             session_data.m_formulas.push_back(
-                new xlsx_session_data::formula(
+                orcus::make_unique<xlsx_session_data::formula>(
                     m_sheet_id, m_cur_row, m_cur_col, m_cur_formula.str.str(), m_cur_formula.ref.str()));
         }
         else
         {
             // normal (non-shared) formula expression
             session_data.m_formulas.push_back(
-                new xlsx_session_data::formula(
+                orcus::make_unique<xlsx_session_data::formula>(
                     m_sheet_id, m_cur_row, m_cur_col, m_cur_formula.str.str()));
         }
     }
-    else if (m_cur_formula.type == spreadsheet::formula_shared && m_cur_formula.shared_id >= 0)
+    else if (m_cur_formula.type == spreadsheet::formula_t::shared && m_cur_formula.shared_id >= 0)
     {
         // shared formula without formula expression
         session_data.m_shared_formulas.push_back(
-            new xlsx_session_data::shared_formula(
+            orcus::make_unique<xlsx_session_data::shared_formula>(
                 m_sheet_id, m_cur_row, m_cur_col, m_cur_formula.shared_id));
     }
-    else if (m_cur_formula.type == spreadsheet::formula_data_table)
+    else if (m_cur_formula.type == spreadsheet::formula_t::data_table)
     {
         // Import data table.
         spreadsheet::iface::import_data_table* dt = mp_sheet->get_data_table();
@@ -596,7 +581,7 @@ void xlsx_sheet_context::end_element_cell()
         {
             if (m_cur_formula.data_table_2d)
             {
-                dt->set_type(spreadsheet::data_table_both);
+                dt->set_type(spreadsheet::data_table_type_t::both);
                 dt->set_range(m_cur_formula.ref.get(), m_cur_formula.ref.size());
                 dt->set_first_reference(
                     m_cur_formula.data_table_ref1.get(), m_cur_formula.data_table_ref1.size(),
@@ -607,7 +592,7 @@ void xlsx_sheet_context::end_element_cell()
             }
             else if (m_cur_formula.data_table_row_based)
             {
-                dt->set_type(spreadsheet::data_table_row);
+                dt->set_type(spreadsheet::data_table_type_t::row);
                 dt->set_range(m_cur_formula.ref.get(), m_cur_formula.ref.size());
                 dt->set_first_reference(
                     m_cur_formula.data_table_ref1.get(), m_cur_formula.data_table_ref1.size(),
@@ -615,7 +600,7 @@ void xlsx_sheet_context::end_element_cell()
             }
             else
             {
-                dt->set_type(spreadsheet::data_table_column);
+                dt->set_type(spreadsheet::data_table_type_t::column);
                 dt->set_range(m_cur_formula.ref.get(), m_cur_formula.ref.size());
                 dt->set_first_reference(
                     m_cur_formula.data_table_ref1.get(), m_cur_formula.data_table_ref1.size(),
@@ -646,21 +631,21 @@ void xlsx_sheet_context::push_raw_cell_value()
 
     switch (m_cur_cell_type)
     {
-        case cell_type_string:
+        case xlsx_ct_shared_string:
         {
             // string cell
             size_t str_id = to_long(m_cur_value);
             mp_sheet->set_string(m_cur_row, m_cur_col, str_id);
         }
         break;
-        case cell_type_value:
+        case xlsx_ct_numeric:
         {
             // value cell
             double val = to_double(m_cur_value);
             mp_sheet->set_value(m_cur_row, m_cur_col, val);
         }
         break;
-        case cell_type_boolean:
+        case xlsx_ct_boolean:
         {
             // boolean cell
             bool val = to_long(m_cur_value) != 0;
@@ -672,6 +657,11 @@ void xlsx_sheet_context::push_raw_cell_value()
     }
 }
 
+void xlsx_sheet_context::pop_rel_extras(opc_rel_extras_t& other)
+{
+    m_rel_extras.swap(other);
+}
+
 }
 
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
